module Lib
    ( polIncPar,
    polIncParChunk, 
    polIncSerial
    ) where

import qualified Data.Vector as V
import Data.List
import Control.Parallel.Strategies

import MapCompat (IntMap)
import qualified Data.IntMap as Map
import Data.Ord (comparing)
import qualified Data.Set as S
import qualified Control.Monad.State as StM
-- import Data.List (sortBy, group, sort)
import Control.Arrow ((&&&))
import qualified Data.List.Split as Split
import qualified Data.ByteString.Lazy as BL
import qualified Data.Csv as CSV
import qualified Data.IntSet as IntSet

type Vertex = Int
type Weight = Int
type Score = Double
type Graph = IntMap (IntMap Weight)

weight :: Graph -> Vertex -> Vertex -> Maybe Weight
weight g i j = do
  jmap <- Map.lookup i g
  Map.lookup j jmap

insertEdge :: Vertex -> Vertex -> Weight -> Graph -> Graph
insertEdge i j w m = Map.insert i iarr m'
          where iarr = Map.insert j w ia
                ia = Map.findWithDefault Map.empty i m'
                m' = Map.insert j jarr m
                jarr = Map.insert i w ja
                ja = Map.findWithDefault Map.empty j m

readGraph :: String -> IO (V.Vector (Int, Int))
readGraph fileName = do
    let f = fileName
    -- let f = "/Users/saiteja/Documents/Columbia/Functional-Programming/playground/PolInc/git_web_ml/salaries.csv"
    -- let f = "/Users/saiteja/Documents/Columbia/Functional-Programming/playground/PolInc/git_web_ml/musae_git_edges.csv"
    -- let f = "/Users/saiteja/Documents/Columbia/Functional-Programming/playground/PolInc/wiki-Vote.txt"
    -- let f = "/Users/saiteja/Documents/Columbia/Functional-Programming/playground/PolInc/soc-wiki-Vote/soc-wiki-Vote.mtx"
    -- let f = "/Users/saiteja/Documents/Columbia/Functional-Programming/playground/PolInc/soc-hamsterster/soc-hamsterster.edges"
    BL.readFile f >>= dat . (CSV.decode CSV.NoHeader)
    where dat (Left _) = undefined
          dat (Right v) = return $ app v
            where app v' = do
                            (v1 :: Int, v2 :: Int) <- v'
                            return (v1, v2)

getGraph :: [(Vertex, Vertex)] -> (Graph, [Vertex])
getGraph edges = (mat, vers)
    where 
        vers = getVertices edges
        mat = foldr ins Map.empty edges
            where ins (i,j) = insertEdge i j 1

getVertices' :: [(Vertex, Vertex)] -> [Vertex]
getVertices' [] = []
getVertices' ((v1, v2):xs) = [v1, v2] ++ (getVertices' xs)

getVertices :: [(Vertex, Vertex)] -> [Vertex]
getVertices edges = IntSet.elems (IntSet.fromList $ getVertices' edges)

topPercent :: Double -> Int -> Int
topPercent percent count | ((fromIntegral count)*percent/100.0) > 1.0 = ceiling $ (fromIntegral count)*percent/100.0
                 | otherwise = 1

processVertexMap :: Vertex -> [Vertex] -> Graph -> Vertex -> IntMap Weight -> IntMap Weight
processVertexMap k vs g i jmap = foldr shortest Map.empty vs
    where shortest j m =
            case (old,new) of
              (Nothing, Nothing) -> m
              (Nothing, Just w ) -> Map.insert j w m
              (Just w,  Nothing) -> Map.insert j w m
              (Just w1, Just w2) -> Map.insert j (min w1 w2) m
            where
              old = Map.lookup j jmap
              new = do 
                      if (i == j)
                        then do return 0
                      else do
                            w1 <- weight g i k
                            w2 <- weight g k j
                            return (w1+w2)

shortestPathsParallel :: [Vertex] -> Graph -> Graph
shortestPathsParallel vs graph = foldl' update graph vs
 where
  update g k = Map.fromList $ parMap rdeepseq (\(i, jmap) -> (i, (processVertexMap k vs g) i jmap)) (Map.toList g)

shortestPaths :: [Vertex] -> Graph -> Graph
shortestPaths vs graph = foldl' update graph vs 
 where
  update g k = Map.mapWithKey (processVertexMap k vs g) g

degreeMeasures :: Graph -> [(Int, Int)]
degreeMeasures graph = [ (a, length b) | (a, b) <- (Map.toList graph)]

closenessMeasures :: Graph -> [(Int, Int)]
closenessMeasures sPaths = [ (a, sum $ Map.elems b) | (a, b) <- (Map.toList sPaths)]

closenessMeasuresHelper :: Graph -> [(Int, Int)]
closenessMeasuresHelper sp = concat (map (closenessMeasures.(Map.fromList)) chunks `using` parList rdeepseq)
                          where chunks = Split.chunksOf (length sp `quot` 8) (Map.toList sp)

closenessMeasuresHelperNoChunks :: Graph -> [(Int, Int)]
closenessMeasuresHelperNoChunks sPaths = concat (map (closenessMeasures.Map.fromList.(\x -> [x])) (Map.toList sPaths) `using` parList rdeepseq)

countOccurences :: Ord a => [a] -> [(a, Int)]
countOccurences = map (head &&& length) . group . sort

getCounts :: [(Int, Score)] -> [(Int, Int)]
getCounts scores = countOccurences.Map.elems.Map.fromList $ map (\(a, b) -> (a, (\c -> if c >= 0 then 1 else -1) b)) scores

computeSummary :: IntMap Score -> IO ()
computeSummary scores = do
               mapM_ (putStrLn.show) (getCounts $ Map.toList scores)

getNeighbours ::  Int -> Graph -> [Vertex]
getNeighbours source graph = case (Map.lookup source graph) of 
                                 Nothing -> []
                                 Just v -> Map.keys v

diffuse :: Double -> Double -> Double -> Double -> Double
diffuse strength distance diameter alpha = strength * (exp $ (-1.0) * alpha * distance/diameter)

getNextSources :: Int -> [Vertex] -> Graph -> Double -> StM.State (IntMap Score) [Vertex]
getNextSources dist sources graph strength = do
    visited <- StM.get
    let newNeigh = S.fromList $ concat [getNeighbours s graph | s <- sources]
        nextsources = S.toList $ S.difference newNeigh $ S.fromList $ Map.keys visited
        score = diffuse strength (fromIntegral dist) 6 1
    StM.put $ foldr (\a -> Map.insert a score) visited nextsources
    return nextsources

doBFS :: Int -> [Vertex] -> Graph -> Double -> StM.State (IntMap Score) (Graph)
doBFS dist sources graph strength = do
      newsources <- getNextSources dist sources graph strength
      if | length newsources == 0 -> return graph
         | otherwise -> doBFS (dist + 1) newsources graph strength

generateScores:: Graph -> (Vertex, Double) -> IntMap Score
generateScores graph (source, strength) = StM.execState (doBFS 1 [source] graph strength) (Map.fromList [(source, 0)])

getScores:: Graph -> [(Vertex, Double)] -> IntMap Score
getScores graph influencers = foldr (\a b -> Map.unionWith (+) (generateScores graph a) b) Map.empty influencers

getScoresHelper:: Graph -> [(Vertex, Double)] -> Int -> IntMap Score
getScoresHelper graph influencers chunkSize = foldr (\a b -> Map.unionWith (+) a b) Map.empty (map (generateScores graph) influencers `using` parListChunk chunkSize rdeepseq)
                            -- where chunkSize = 50
                            -- where chunkSize = (length influencers) `quot` 37
                            -- where chunkSize = (length influencers) `quot` 5

getScoresHelperNoChunks:: Graph -> [(Vertex, Double)] -> IntMap Score
getScoresHelperNoChunks graph influencers = foldr (\a b -> Map.unionWith (+) a b) Map.empty (map ((getScores graph).(\x -> [x])) influencers `using` parList rdeepseq)

polIncParChunk :: Int -> String -> IO ()
polIncParChunk chunkSize fileName = do
    graph <- readGraph fileName
    let edges = V.toList graph
        (mat,vs) = getGraph edges
        sp = shortestPathsParallel vs mat
        -- deg = degreeMeasures mat
        cm = closenessMeasuresHelper sp
        influencers = map fst $ take (topPercent 20 $ length vs) $ sortBy (comparing snd) $ cm
        -- influencers = map fst $ take (topPercent 10 $ length vs) $ reverse $ sortBy (comparing snd) $ deg
    putStrLn $ show $ length vs
    putStrLn "polIncParChunk running" 
    let netscores = getScoresHelper mat (zip influencers (cycle [1, -1])) chunkSize
    computeSummary netscores

polIncPar :: String -> IO ()
polIncPar fileName = do
    graph <- readGraph fileName
    let edges = V.toList graph
        (mat,vs) = getGraph edges
        sp = shortestPathsParallel vs mat
        -- deg = degreeMeasures mat
        cm = closenessMeasuresHelperNoChunks sp
        influencers = map fst $ take (topPercent 20 $ length vs) $ sortBy (comparing snd) $ cm
        -- influencers = map fst $ take (topPercent 10 $ length vs) $ reverse $ sortBy (comparing snd) $ deg
    putStrLn $ show $ length vs
    putStrLn "polIncPar running" 
    let netscores = getScoresHelperNoChunks mat (zip influencers (cycle [1, -1]))
    computeSummary netscores

polIncSerial :: String -> IO ()
polIncSerial fileName = do
    graph <- readGraph fileName
    let edges = V.toList graph
        (mat,vs) = getGraph edges
        -- sp = shortestPaths vs mat
        deg = degreeMeasures mat
        -- cm = closenessMeasures sp
        -- influencers = map fst $ take (topPercent 20 $ length vs) $ sortBy (comparing snd) $ cm
        influencers = map fst $ take (topPercent 10 $ length vs) $ reverse $ sortBy (comparing snd) $ deg
    putStrLn $ show $ length vs
    putStrLn "polIncSerial running" 
    let netscores = getScores mat (zip influencers (cycle [1, -1]))
    computeSummary netscores
